/*
 File: GHexomino_prototype.h
 GHexomino prototype.

 This is a toy (non-optimized) implementation of the importance sampling
 technique proposed in the paper:
 "Polyomino-Based Digital Halfotning", by Victor Ostromoukhov,
 to be presented at SIGGRAPH 2007.
 
 Implementation by Victor Ostromoukhov
 Universite de Montreal 2007

*/


#ifndef GHexomino_PROTOTYPE_H
#define GHexomino_PROTOTYPE_H

#include <math.h>
#include <vector>

#define MIN(x,y) ((x)<(y)?(x):(y))
#define MAX(x,y) ((x)>(y)?(x):(y))

#define SCALE_RATIO 9
#define AREA_SCALE_RATIO 81
#define NTYPES 436

#define LOG81	4.39444915467244

/// Simple 2D point and vector type.
class Point2D
{
public:
  double x,y;

  Point2D(){};
  Point2D(const double x, const double y) { this->x=x; this->y=y; }
  Point2D(const double vect[2]) { x=vect[0]; y=vect[1]; }

  Point2D operator+(const Point2D& pt) const{ return Point2D(x+pt.x,y+pt.y); }
  Point2D operator-(const Point2D& pt) const{ return Point2D(x-pt.x,y-pt.y); }
  Point2D operator*(double factor) const{ return Point2D(x*factor,y*factor); }
  Point2D operator/(double factor) const{ return Point2D(x/factor,y/factor); }

  /// Returns the squared distance to the origin, or the squared length of a vector.
  double d2() const { return x*x+y*y; }
};

/// This is a pure-virtual class, and you must implement the "getImportanceAt()" function
/// in order to use the sampling system.
/// The mechanics of the system can be observed in the given source code.
class GHexomino
{

protected:

  //
  // Static tables.
  //


	// octavetab: powers of AREA_SCALE_RATIO
	static const double octavetab[11];

  /// 8 basic vectors v1 and v2
  static const Point2D eightbasesv1[8]; // defined at end of file.
  static const Point2D eightbasesv2[8]; // defined at end of file.

	/// inflation parameters: xy and dir
  static const int inflation_xy [AREA_SCALE_RATIO][2]; // defined at end of file.
	static const int inflation_dir [AREA_SCALE_RATIO]; // defined at end of file.

  /// Pre-calculated correction vectors lookup table.
  /// These are available in ASCII format on the web-site.
  static const double lut[AREA_SCALE_RATIO-1][AREA_SCALE_RATIO*NTYPES][2]; // defined at end of file.

  static const int ranking [NTYPES][AREA_SCALE_RATIO]; // defined at end of file.
  static const int productionRules [NTYPES][AREA_SCALE_RATIO]; // defined at end of file.


  /// Returns the required level of subdivision for a given importance value.
  static unsigned getReqSubdivisionLevel( unsigned importance )
  {
    unsigned res = 0;
    if (importance==0) return res;
    res = (unsigned)ceil(log((double)importance)/LOG81 );
    return res;
  }  



  //
  // Inner classes.
  //


  /// Individual tile elements, which also serve as nodes for the tile subdivision tree.
  class TileNode
  {

    unsigned level; // Depth in the tree.
    int tileType;
    int dir; // 8 possible tile orientationis, given by eightbasesv1 and eightbasesv2
    double scale; 
    Point2D refPt; 
    Point2D v1, v2; //
    unsigned threshold; 
    int structuralIndex;
    
    // tiling tree structure
    TileNode* parent;
    unsigned parent_slot; // position in parent's list (needed for iterators)
    bool terminal; // true for leaf nodes
    std::vector<TileNode*> children;

  public:

    /// Builds a tile according to the given specifications.
    TileNode(  
      TileNode* parent = NULL, 
      int tileType = 1,
      Point2D refPt = Point2D(0,0),
      Point2D v1 = Point2D(1,0),
      Point2D v2 = Point2D(0,1),
			unsigned threshold = 0,
			int structuralIndex = 0,
      int parent_slot = 0,
      double scale = 1.0)
    {
      this->parent = parent;
      this->tileType = tileType;
      this->refPt = refPt;
      this->v1 = v1;
      this->v2 = v2;
      this->threshold = threshold;
      this->structuralIndex = structuralIndex;
      this->parent_slot = parent_slot;
      this->scale = scale;
      this->level = parent ? parent->level + 1 : 0; // Increment the level.

      terminal = true;
      children.clear();
    }

    /// Helper constructor.
    /// Creates an initial tile that is certain to contain the ROI.
    /// The starting tile is of type F (arbitrary).
    TileNode( double roi_width,  double roi_height)
    {
      double side = MAX(roi_width,roi_height);
      double scale = 1.0;
      Point2D offset(0, 0);
      *this = TileNode(NULL, 1,offset,Point2D(1,0),Point2D(0,1),0,0,0,scale);
    }
    
    ~TileNode() { collapse(); }

    /// Splits a tile according to the given subdivision rules.
    /// Please refer to the code for further details.
    void refine()
    {
      int i;
      if (!terminal) return; // Can only subdivide leaf nodes.
      
      terminal=false; // The tile now has children.
      double newscale = scale / SCALE_RATIO; // The scale factor between levels is constant.
			int scaleint = (int)round(1./(scale*scale));
      
      // Each new tile is created using the following information:
      // A pointer to its parent, the type of the new tile
      // the origin of the new tile, the change in orientation of the new tile with
      // respect to the parent's orientation
      // the parent's slot (for traversal purposes), and the new linear scale of the tile,
      // which is always the parent's scale divided by the SCALE_RATIO.

      for (unsigned i=0; i<AREA_SCALE_RATIO; i++) {
        Point2D newrefPt;
        Point2D newv1, newv2;
        unsigned newthreshold;
        int newstructuralIndex;

        int thisTileRanking = ranking[tileType-1][i];
				int newx = inflation_xy[thisTileRanking-1][0]; 
				int newy = inflation_xy[thisTileRanking-1][1]; 
				int newdir = inflation_dir[thisTileRanking-1]; 
				int newTileType = productionRules[tileType-1][thisTileRanking-1]; 
				Point2D newbasev1 = eightbasesv1[newdir];
				Point2D newbasev2 = eightbasesv2[newdir];
				
				newv1 = Point2D(newbasev1.x*v1.x + newbasev1.y*v2.x, newbasev1.x*v1.y + newbasev1.y*v2.y);
				newv2 = Point2D(newbasev2.x*v1.x + newbasev2.y*v2.x, newbasev2.x*v1.y + newbasev2.y*v2.y);
				newrefPt = refPt + Point2D((newx*v1.x + newy*v2.x)*newscale, (newx*v1.y + newy*v2.y)*newscale);
				newstructuralIndex = (tileType-1)*AREA_SCALE_RATIO + i;
        newthreshold = threshold + i * scaleint;
        children.push_back(
          new TileNode(this, newTileType, newrefPt, newv1, newv2, newthreshold, newstructuralIndex, i, newscale) );
      }
    }

    /// Prunes the subdivision tree at this node.
    void collapse()
    {
      // Recursively prune the tree.
      for (unsigned i=0; i<children.size(); i++) delete children[i];
      terminal = true;
      children.clear();
    }

    /// Returns the next node of the tree, in depth-first traversal.
    /// Returns NULL if it is at the last node.
    TileNode* nextNode()
    {
      if (!terminal) return children[0];
      
      if (level == 0) return NULL; // single node case.

      if ( parent_slot < parent->children.size()-1 )
        return parent->children[parent_slot+1];

      // last child case
      TileNode* tmp = this;
      do
      {
        tmp = tmp->parent;
      }
      while ( (tmp->level != 0) && (tmp->parent_slot ==  tmp->parent->children.size()-1) );

      if (tmp->level == 0) return NULL; // last node
      return tmp->parent->children[tmp->parent_slot+1];

    }

    /// Returns the next closest leaf to a node.
    /// Returns NULL if it's the last leaf.
    TileNode* nextLeaf()
    {
      TileNode* tmp = this;
      do
      {
        tmp = tmp->nextNode();
        if ( !tmp ) return NULL;
        if ( tmp->terminal ) return tmp;
      }
      while (1);
    }

    // Public accessors

    Point2D getrefPt() const { return refPt; }
    unsigned getthreshold() const { return threshold; }
    unsigned getLevel() { return level; }
    bool isTerminal() const { return terminal; }
    TileNode* getParent() { return parent; }
    TileNode* getChild(unsigned i) {  return children[i]; }

    /// Obtains the correction vector from the lookup table,
    /// then scales and adds it to the reference point.
    Point2D getDisplacedSamplingPoint(unsigned importance)
    {
			int importanceIndex = (int)ceil((double)(AREA_SCALE_RATIO-1)*(importance-octavetab[level-1])/(octavetab[level]-octavetab[level-1]));
			Point2D lutvalue = lut[importanceIndex][structuralIndex];
			Point2D res = refPt + (v1 + v2 + Point2D(lutvalue.x*v1.x + lutvalue.y*v2.x, lutvalue.x*v1.y + lutvalue.y*v2.y) ) * scale;
			return res;
    }  

  }; // end of class TileNode.

  /// Leaf iterator for the tile subdivision tree.
  /// The traversal is made in a depth-first manner.
  /// Warning: This does not behave like STL style iterators.
  class TileLeafIterator
  {
    TileNode* shape;
  public:
    TileLeafIterator() { shape=NULL; }
    TileLeafIterator(TileNode* s ) { begin(s); }

    TileNode* operator*() { return shape; }
    TileNode* operator->() { return shape; }

    void begin(TileNode* s)
    {
      TileNode* tmp = s;  
      while ( ! tmp->isTerminal() ) tmp = tmp->getChild(0); // find first leaf
      shape = tmp;
    }

    /// Subdivides the tile and moves to its 1st child.
    void refine() 
    {
      shape->refine();
      shape = shape->getChild(0);
    }

    /// Prunes the subdivision tree.
    void collapse()
    {
      if (shape->getParent())
      {
        shape = shape->getParent();
        shape->collapse();
      }
    }

    /// Moves to the next node in the subdivision tree, in depth-first traversal.
    /// Returns false iff there is no such node.
    bool next()
    {
      TileNode* s = shape->nextLeaf();
      if (s)
      {
        shape = s;
        return true;
      }
      else
      {
        shape = s;
        return false;
      }
    }

    /// Checks if there is a next tile, in depth-first traversal.
    bool hasNext()
    {
      TileNode* s = shape->nextLeaf();
      if (s) return true;
      else return false;
    }
  };


  //
  // Instance members.
  //

  /// Root node of the tile subdivision tree.
  TileNode *root;
  
  /// Extents of the region of interest.
  double width, height;

  /// Protected constructor, which initializes the Region of Interest.
  GHexomino(double width=0.0, double height=0.0) 
  { this->width=width; this->height=height; root=NULL; }
  
  virtual ~GHexomino() { if (root) delete root; }



  /// This is a helper function which constrains the incoming points
  /// to the region of interest.
  unsigned getImportanceAt_bounded(Point2D pt)
  {
    if (pt.x>=0 && pt.x<width && pt.y>=0 && pt.y<height) 
      return getImportanceAt(pt);
    else 
      return 0;
  }

  /// Subdivides all tiles down a level, a given number of times.
  void subdivideAll(int times=1)
  {
    if (!root) return;
    TileNode *tmp;
    for (int i=0;i<times;i++)  
    {
      TileLeafIterator it(root);
      do { 
        tmp = *it;
        it.next();
        tmp->refine();
      }
      while (*it);
    }
  }

  /// Generates the hierarchical structure.
  void buildAdaptiveSubdivision( unsigned minSubdivisionLevel = 1 )
  {
    root = new TileNode(width,height);
    
    // Since we are approximating the MAX within each tile by the values at
    // a few key points, we must provide a sufficiently dense initial
    // tiling. This would not be necessary with a more thorough scan of each
    // tile.
    subdivideAll(minSubdivisionLevel);

    TileLeafIterator it(root);
    TileNode *tmp;

    // Recursively subdivide all triangles until each triangle's
    // required level is reached.
    unsigned level;
    do {
      level = it->getLevel();
      
      {
        if ( level < getReqSubdivisionLevel(getImportanceAt_bounded(it->getrefPt())) )
        {
          tmp = *it;
          tmp->refine();
        }
      }
    } while ( it.next() );
  }


  /// Collect the resulting point set.
  void collectPoints( 
    std::vector<Point2D> &pointlist,
    bool filterBounds = false )
  {
    pointlist.clear();

    Point2D pt, pt_displaced;
    unsigned importance;
    TileLeafIterator it(root);
    do {
      pt = it->getrefPt();  
        importance = getImportanceAt_bounded( pt );
      // xxx
        if ( importance >  it->getthreshold() )
        {
          // Get the displaced point using the lookup table.
          pt_displaced = it->getDisplacedSamplingPoint(importance);

          if ( !filterBounds ||
            (pt_displaced.x>=0 && pt_displaced.x<width && 
            pt_displaced.y>=0 && pt_displaced.y<height) )
          {
            pointlist.push_back(pt_displaced); // collect point.
          }
        }

    } while ( it.next() );

  }



public:

  /// This virtual function must be implemented in order to use the sampling system.
  /// It should return the value of the importance function at the given point.
  virtual unsigned getImportanceAt( Point2D pt ) = 0;

  /// Builds and collects the point set generated be the sampling system,
  /// using the previously defined importance function.
  std::vector<Point2D> getSamplingPoints()
  {
    if (root) delete root;
    std::vector<Point2D> pointlist;

    buildAdaptiveSubdivision();
    collectPoints(pointlist);
    return pointlist;
  }

  /// \example example.cpp
  /// This a simple example of how to use the GHexomino class.

};



/*

Static Member initialization

*/


const double GHexomino::octavetab[] = 
{1.,81.,6561.,531441.,4.3046721e7,3.486784401e9,2.82429536481e11,2.2876792454961e13,1.85302018885184e15,1.50094635296999e17,1.21576654590569e19};
const unsigned int octavetabInt[] = {1,81,6561,531441,43046721,3486784401};


const Point2D GHexomino::eightbasesv1[8] = {
  Point2D( 1, 0),
  Point2D( 0, 1),
  Point2D(-1, 0),
  Point2D( 0,-1),
  Point2D( 0, 1),
  Point2D(-1, 0),
  Point2D( 0,-1),
  Point2D( 1, 0)
};

const Point2D GHexomino::eightbasesv2[8] = {
  Point2D( 0, 1),
  Point2D(-1, 0),
  Point2D( 0,-1),
  Point2D( 1, 0),
  Point2D( 1, 0),
  Point2D( 0, 1),
  Point2D(-1, 0),
  Point2D( 0,-1)
};

const int GHexomino::inflation_xy [AREA_SCALE_RATIO][2] =
{{0, 0}, {5, 0}, {7, 3}, {9, 0}, {9, 0}, {14, 0}, {16, 3}, {18,
  0}, {3, 2}, {3, 3}, {5, 6}, {7, 5}, {12, 2}, {12, 3}, {14, 4}, {18,
  5}, {18, 8}, {0, 4}, {3, 6}, {3, 8}, {6, 9}, {10, 5}, {10, 7}, {13,
  9}, {13, 9}, {0, 8}, {3, 10}, {3, 12}, {6, 10}, {6, 12}, {10,
  8}, {10, 11}, {13, 9}, {18, 8}, {0, 12}, {6, 14}, {9, 14}, {9,
  11}, {14, 13}, {17, 13}, {17, 12}, {19, 9}, {22, 9}, {23, 13}, {24,
  11}, {27, 9}, {4, 16}, {7, 16}, {10, 16}, {13, 13}, {13, 13}, {18,
  15}, {18, 12}, {24, 15}, {27, 13}, {0, 15}, {3, 18}, {8, 18}, {11,
  18}, {11, 18}, {16, 18}, {18, 15}, {20, 18}, {23, 18}, {19,
  14}, {27, 18}, {25, 15}, {0, 19}, {3, 17}, {3, 18}, {5, 19}, {9,
  20}, {9, 23}, {0, 22}, {0, 27}, {2, 24}, {4, 27}, {4, 23}, {6,
  24}, {6, 25}, {9, 27}};

const int GHexomino::inflation_dir [AREA_SCALE_RATIO] =
{0, 1, 6, 1, 0, 1, 6, 1, 5, 4, 3, 3, 5, 4, 0, 2, 2, 0, 5, 7, 7, 5, 7,
7, 2, 0, 5, 7, 2, 7, 5, 7, 4, 1, 0, 2, 2, 4, 6, 6, 3, 4, 4, 6, 0, 5,
2, 2, 2, 5, 0, 2, 0, 2, 5, 0, 3, 2, 2, 7, 6, 1, 6, 6, 4, 6, 1, 0, 5,
4, 0, 2, 2, 0, 3, 4, 3, 2, 6, 7, 2};

const double GHexomino::lut[AREA_SCALE_RATIO-1][AREA_SCALE_RATIO*NTYPES][2] = 
#include "G-hexomino_rep-81_lut.data"

const int GHexomino::ranking [NTYPES][AREA_SCALE_RATIO] =
#include "G-hexomino_rep-81_ranking.data"

const int GHexomino::productionRules [NTYPES][AREA_SCALE_RATIO] =
#include "G-hexomino_rep-81_productionRules.data"

#endif //GHexomino_PROTOTYPE_H


